# -*- coding: utf-8 -*-
import time

from pip_services4_components.exec.FixedRateTimer import FixedRateTimer

class TestFixedRateTimer:
    counter = 0

    def setup_method(self):
        self.counter = 0

    def test_run_with_task(self):

        def callback():
            self.counter += 1

        timer = FixedRateTimer(callback, 100, 0)

        timer.start()
        time.sleep(0.5)
        timer.stop()

        assert self.counter > 3

    def test_run_with_callback(self):
        def callback():
            self.counter += 1

        timer = FixedRateTimer(callback, 100, 0)

        timer.start()
        time.sleep(0.5)
        timer.stop()

        assert self.counter > 3



# -*- coding: utf-8 -*-
"""
    pip_services4_components.build.CreateException
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Create exception type
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Optional

from pip_services4_commons.errors.InternalException import InternalException


class CreateException(InternalException):
    """
    Error raised when factory is not able to create requested component.
    """

    def __init__(self, trace_id: Optional[str] = None, message_or_locator: str = None):
        """
        Creates an error instance and assigns its values.

        :param trace_id: (optional) a unique transaction id to trace execution through call chain.

        :param message_or_locator: human-readable error or locator of the component that cannot be created.
        """
        super(CreateException, self).__init__(trace_id, "CANNOT_CREATE", message_or_locator)



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.refer.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Inversion of control design pattern. There exist various implementations,
    a popular one being "inversion of dependency". Requires introspection and
    is implemented differently in different languages. In PipServices, the "location
    design pattern” is used, which is much simpler than dependency injection and is
    a simple implementation, that is portable between languages. Used for building
    various containers, as well as testing objects.

    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = [
    'Descriptor',
    'IReferenceable', 'IUnreferenceable', 'IReferences',
    'ReferenceException', 'Referencer', 'Reference',
    'References', 'DependencyResolver'
]

from .DependencyResolver import DependencyResolver
from .Descriptor import Descriptor
from .IReferenceable import IReferenceable
from .IReferences import IReferences
from .IUnreferenceable import IUnreferenceable
from .Reference import Reference
from .ReferenceException import ReferenceException
from .Referencer import Referencer
from .References import References



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.exec.Cleaner
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Cleaner component implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Any, List, Optional

from pip_services4_components.context.IContext import IContext

from .ICleanable import ICleanable


class Cleaner:
    """
    Helper class that cleans stored object state.
    """

    @staticmethod
    def clear_one(context: Optional[IContext], component: Any):
        """
        Clears state of specific component.

        To be cleaned state components must implement :class:`ICleanable <pip_services4_components.run.ICleanable.ICleanable>` interface.
        If they don't the call to this method has no effect.

        :param context: (optional) transaction id to trace execution through call chain.

        :param component: the component that is to be cleaned.

        """
        if isinstance(component, ICleanable):
            component.clear(context)

    @staticmethod
    def clear(context: Optional[IContext], components: List[Any]):
        """
        Clears state of multiple components.

        To be cleaned state components must implement :class:`ICleanable <pip_services4_components.run.ICleanable.ICleanable>` interface.
        If they don't the call to this method has no effect.

        :param context: (optional) transaction id to trace execution through call chain.

        :param components: the list of components that are to be cleaned.
        """
        if components is None:
            return

        for component in components:
            Cleaner.clear_one(context, component)



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.exec.FixedRateTimer
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Fixed rate timer implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
import inspect
import time
from threading import Thread, Event, Lock
from typing import Callable, Any, Optional

from pip_services4_components.context import Context
from pip_services4_components.context.IContext import IContext

from pip_services4_components.run.IClosable import IClosable
from pip_services4_components.exec.INotifiable import INotifiable
from pip_services4_components.exec.Parameters import Parameters


class Timer(Thread):
    def __init__(self, interval, delay, callback):
        Thread.__init__(self)
        self._interval = interval
        self._callback = callback
        self._event = Event()
        self._delay = delay

    def run(self):
        time.sleep(self._delay)
        while not self._event.is_set():
            self._callback()
            time.sleep(self._interval)

    def stop(self):
        if self.is_alive():
            # set event_name to signal thread to terminate
            self._event.set()
            # block calling thread until thread really has terminated
            self.join()


class FixedRateTimer(IClosable):
    """
    Timer that is triggered in equal time intervals.

    It has summetric cross-language implementation
    and is often used by Pip.Services toolkit to perform periodic processing and cleanup in microservices.
    """

    __lock = None

    def __init__(self, task_or_object: Any = None, interval: int = None, delay: int = None):
        """
        Creates new instance of the timer and sets its values.

        :param task_or_object: (optional) a Notifiable object or callback function to call when timer is triggered.

        :param interval: (optional) an __interval to trigger timer in milliseconds.

        :param delay: (optional) a __delay before the first triggering in milliseconds.
        """
        self.__lock = Lock()
        self.__task: INotifiable = None
        self.__callback: Callable = None
        self.__delay: int = None
        self.__interval: int = None
        self.__timer: Any = None
        self.__started: bool = False

        if hasattr(task_or_object, 'notify') and inspect.ismethod(
                task_or_object.notify):
            self.set_task(task_or_object)
        else:
            self.set_callback(task_or_object)

        self.set_interval(interval)
        self.set_delay(delay)

    def get_task(self) -> INotifiable:
        """
        Gets the INotifiable object that receives notifications from this timer.

        :return: the INotifiable object or null if it is not set.
        """
        return self.__task

    def set_task(self, value: INotifiable):
        self.__task = value
        self.__callback = self.__timer_callback

    def get_callback(self) -> Callable:
        """
        Gets the callback function that is called when this timer is triggered.

        :return: the callback function or null if it is not set.
        """
        return self.__callback

    def set_callback(self, value: Callable):
        """
        Sets the callback function that is called when this timer is triggered.

        :param value: the callback function to be called.
        """
        self.__callback = value
        self.__task = None

    def get_delay(self) -> int:
        """
        Gets initial delay before the timer is triggered for the first time.

        :return: the delay in milliseconds.
        """
        return self.__delay

    def set_delay(self, value: int):
        """
        Sets initial delay before the timer is triggered for the first time.
        :param value: a delay in milliseconds.
        """
        self.__delay = value

    def get_interval(self) -> int:
        """
        Gets periodic timer triggering interval.

        :return: the interval in milliseconds
        """
        return self.__interval

    def set_interval(self, value: int):
        """
        Sets periodic timer triggering interval.

        :param value: an interval in milliseconds.
        """
        self.__interval = value

    def is_started(self) -> bool:
        """
        Checks if the timer is started.

        :return: true if the timer is started and false if it is stopped.
        """
        return self.__timer is not None

    def start(self):
        """
        Starts the timer.
        Initially the timer is triggered after __delay.
        After that it is triggered after __interval until it is stopped.
        """
        self.__lock.acquire()
        try:
            # Stop previously set timer
            if not (self.__timer is None):
                self.__timer.stop()
                self.__timer = None

            if self.__interval is None or self.__interval <= 0:
                return

            delay = max(0, self.__delay - self.__interval)

            # Set a new timer
            self.__timer = Timer(self.__interval / 1000, delay / 1000, self.__callback)
            self.__timer.start()

            # Set __started flag
            self.__started = True
        finally:
            self.__lock.release()

    def __timer_callback(self):
        try:
            self.__task.notify(Context.from_trace_id("pip-commons-timer"), Parameters())
        except:
            # Ignore or better log
            pass

    def stop(self):
        """
        Stops the timer.
        """
        self.__lock.acquire()
        try:
            # Stop the timer
            if not (self.__timer is None):
                self.__timer.stop()
                self.__timer = None

            # Unset __started flag
            self.__started = False
        finally:
            self.__lock.release()

    def close(self, context: Optional[IContext]):
        """
        Closes the timer.
        This is __required by :class:`IClosable <pip_services4_components.run.IClosable.IClosable>` interface,
        but besides that it is identical to stop().

        :param context: (optional) transaction id to trace execution through call chain.
        """
        self.stop()



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.exec.IExecutable
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Interface for executable components with parameters
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from abc import ABC
from typing import Optional

from pip_services4_components.context.IContext import IContext

from ..exec.Parameters import Parameters


class IExecutable(ABC):
    """
    Interface for components that can be called to execute work.

    .. code-block:: python
        class EchoComponent(IExecutable):
            ...
            def execute(self, context: Optional[IContext], args: Parameters):
                result = args.get_as_object("message")
                return result

        echo = new EchoComponent()
        message = "Test";
        result = echo.execute("123", Parameters.from_tuples("message", message))
        print("Request: " + message + " Response: " + result)

    """

    def execute(self, context: Optional[IContext], args: Parameters):
        """
        Executes component with arguments and receives execution result.

        :param context: (optional) transaction id to trace execution through call chain.

        :param args: execution arguments.

        :return: execution result
        """
        raise NotImplementedError('Method from interface definition')



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.config.NameResolver
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Name resolver implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from .ConfigParams import ConfigParams
from ..refer import Descriptor


class NameResolver:
    """
    A helper class that allows to extract component name from configuration parameters.
    The name can be defined in "id", "name" parameters or inside a component descriptor.

    Examples:

    .. code-block:: python

        config = ConfigParams.fromTuples("descriptor", "myservice:connector:aws:connector1:1.0",
                                         "param1", "ABC",
                                         "param2", 123)

        name = NameResolver.resolve(config)
    """

    @staticmethod
    def resolve(config: ConfigParams, default_name: str = None) -> str:
        """
        Resolves a component name from configuration parameters.
        The name can be stored in "id", "name" fields or inside a component descriptor.
        If name cannot be determined it returns a defaultName.

        :param config: configuration parameters that may contain a component name.

        :param default_name: (optional) a default component name.

        :return: resolved name or default name if the name cannot be determined.
        """
        name = config.get_as_nullable_string("name")
        name = name if not (name is None) else config.get_as_nullable_string("id")

        if name is None:
            descriptor_str = config.get_as_nullable_string("descriptor")
            descriptor = Descriptor.from_string(descriptor_str)
            if not (descriptor is None):
                name = descriptor.get_name()

        return name if not (name is None) else default_name



# -*- coding: utf-8 -*-
"""
    pip_services_common.refer.ReferenceException
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Reference error type
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Any, Optional

from pip_services4_commons.errors import InternalException

from ..context.ContextResolver import ContextResolver
from ..context.IContext import IContext


class ReferenceException(InternalException):
    """
    Error when __required component dependency cannot be found.
    """

    def __init__(self, context: Optional[IContext] = None, locator: Any = None):
        """
        Creates an error instance and assigns its values.

        :param context: (optional) a unique transaction id to trace execution through call chain.

        :param locator: the locator to find reference to dependent component.
        """
        message = 'Cannot locate reference: ' + (str(locator) if not (locator is None) else '<None>')
        super(ReferenceException, self).__init__(ContextResolver.get_trace_id(context), "REF_ERROR", message)
        self.with_details('locator', locator)



from typing import Any

from pip_services4_commons.data import AnyValueMap

from pip_services4_components.config import ConfigParams
from pip_services4_components.context import IContext
from ..exec.Parameters import Parameters


class Context(IContext):
    """
    Basic implementation of an execution context.

    See :class:`IContext <pip_services4_components.context.IContext.IContext>`
    See :class:`AnyValueMap <pip_services4_commons.data.AnyValueMap.AnyValueMap>`
    """

    def __init__(self, values: Any):
        """
        Gets a map element specified by its key.

        :param values: a key of the element to get.
        :return: the value of the map element.
        """
        self._values = AnyValueMap(values)

    def get(self, key: str) -> Any:
        return self._values.get(key)

    @staticmethod
    def from_value(value: Any) -> 'Context':
        """
        Creates a new instance of the map and assigns its value.

        :param value: (optional) values to initialize this map.
        """
        return Context(value)

    @staticmethod
    def from_tuples(*tuples: Any) -> 'Context':
        """
        Creates a new Context object filled with provided key-value pairs called tuples.
        Tuples parameters contain a sequence of key1, value1, key2, value2, ... pairs.

        :param tuples: the tuples to fill a new Parameters object.
        :return: a new Parameters object.

        """

        obj = AnyValueMap.from_tuples(*tuples)
        return Context(obj)

    @staticmethod
    def from_config(config: ConfigParams) -> 'Context':
        """
        Creates new Context from ConfigMap object.

        :param config: a ConfigParams that contain parameters.
        :return: a new Context object.

        See :class:`ConfigParams <pip_services4_components.config.ConfigParams.ConfigParams>`
        """
        if not config:
            return Context()

        values = AnyValueMap()
        for key in config:
            if config.get(key):
                values.put(key, config.get(key))

        return Context(values)

    @staticmethod
    def from_trace_id(trace_id: str) -> 'Context':
        """
        Creates new Context from trace id.
        :param trace_id: a transaction id to trace execution through call chain.
        :return: a new Parameters object.
        """
        return Context(Parameters.from_tuples("trace_id", trace_id))



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.refer.IUnreferenceable
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Interface for unreferenceable components.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from abc import ABC


class IUnreferenceable(ABC):
    """
    Interface for components that require explicit clearing of references to dependent components.

    Example:

    .. code-block:: python

        class MyController(IReferenceable):
            _persistence = None

            def set_references(self, references):
                self._persistence = references.getOneRequired(Descriptor("mygroup", "persistence", "*", "*", "1.0"))

            def unset_references(self):
                self._persistence = None

            
    """

    def unset_references(self):
        """
        Unsets (clears) previously set references to dependent components.
        """
        raise NotImplementedError('Method from interface definition')



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.config.OptionsResolver
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Options resolver implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from ..config.ConfigParams import ConfigParams


class OptionsResolver:
    """
    A helper class to parameters from "options" configuration section.

    Example:

    .. code-block:: python

        config = ConfigParams.fromTuples(
          ...
          "options.param1", "ABC",
          "options.param2", 123)

        options = OptionsResolver.resolve(config)
    """

    @staticmethod
    def resolve(config: ConfigParams, config_as_default: bool = False) -> ConfigParams:
        """
        Resolves an "options" configuration section from component configuration parameters.

        :param config: configuration parameters

        :param config_as_default: (optional) When set true the method returns the entire parameter
                                  set when "options" section is not found. Default: false

        :return: configuration parameters from "options" section
        """
        options = config.get_section("options")

        if len(options) == 0 and config_as_default:
            options = config

        return options



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.refer.Referencer
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Referencer component implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Any, List

from ..refer import IReferences
from .IReferenceable import IReferenceable
from .IUnreferenceable import IUnreferenceable


class Referencer:
    """
    Helper class that sets and unsets references to components.
    """

    @staticmethod
    def set_references_for_one(references: IReferences, component: Any):
        """
        Sets references to specific component.

        To set references components must implement :class:`IReferenceable <pip_services4_commons.refer.IReferenceable.IReferenceable>` interface.
        If they don't the call to this method has no effect.

        :param references: the references to be set.

        :param component: the component to set references to.
        """
        if isinstance(component, IReferenceable):
            component.set_references(references)

    @staticmethod
    def set_references(references: IReferences, components: List[Any]):
        """
        Sets references to multiple components.

        To set references components must implement :class:`IReferenceable <pip_services4_commons.refer.IReferenceable.IReferenceable>` interface.
        If they don't the call to this method has no effect.

        :param references: the references to be set.

        :param components: a list of components to set the references to.
        """
        if components is None:
            return

        for component in components:
            Referencer.set_references_for_one(references, component)

    @staticmethod
    def unset_references_for_one(component: Any):
        """
        Unsets references in specific component.

        To unset references components must implement :class:`IUnreferenceable <pip_services4_compoents.refer.IUnreferenceable.IUnreferenceable>` interface.
        If they don't the call to this method has no effect.

        :param component: the component to unset references.
        """
        if isinstance(component, IUnreferenceable):
            component.unset_references()

    @staticmethod
    def unset_references(components: List[Any]):
        """
        Unsets references in multiple components.

        To unset references components must implement :class:`IUnreferenceable <pip_services4_compoents.refer.IUnreferenceable.IUnreferenceable>` interface.
        If they don't the call to this method has no effect.

        :param components: the list of components, whose references must be cleared.
        """
        if components is None:
            return

        for component in components:
            Referencer.unset_references_for_one(component)



from abc import ABC
from typing import Any


class IContext(ABC):
    """
    Interface to specify execution context.
    """

    def get(self, key: str) -> Any:
        """
        Gets a map element specified by its key.

        :param key: a key of the element to get.
        :return: the value of the map element.
        """
        raise NotImplemented('Method not implemented')



# -*- coding: utf-8 -*-
"""
    pip_services4_components.build.CompositeFactory
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Composite factory implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import List, Any

from .CreateException import CreateException
from .IFactory import IFactory


class CompositeFactory(IFactory):
    """
    Aggregates multiple factories into a single factory component.
    When a new component is requested, it iterates through
    factories to locate the one able to create the requested component.

    This component is used to conveniently keep all supported factories in a single place.

    Example:

    .. code-block:: python
    
        factory = CompositeFactory()
        factory.add(new DefaultLoggerFactory())
        factory.add(new DefaultCountersFactory())

        loggerLocator = Descriptor("*", "logger", "*", "*", "1.0")
        factory.can_create(loggerLocator)  # Result: Descriptor("pip-service", "logger", "null", "default", "1.0")
        factory.create(loggerLocator)      # Result: created NullLogger
    """

    def __init__(self, *factories: IFactory):
        """
        Creates a new instance of the factory.

        :param factories: a list of factories to embed into this factory.
        """
        self.__factories: List[IFactory] = []
        for factory in factories:
            self.__factories.append(factory)

    def add(self, factory: IFactory):
        """
        Adds a factory into the list of embedded factories.

        :param factory: a factory to be added.
        """
        if factory is None:
            raise Exception("Factory cannot be null")

        self.__factories.append(factory)

    def remove(self, factory: IFactory):
        """
        Removes a factory from the list of embedded factories.

        :param factory: the factory to remove.
        """
        self.__factories.remove(factory)

    def can_create(self, locator: Any) -> Any:
        """
        Checks if this factory is able to create component by given locator.

        This method searches for all registered components and returns
        a locator for component it is able to create that matches the given locator.
        If the factory is not able to create a requested component is returns null.

        :param locator: a locator to identify component to be created.

        :return: a locator for a component that the factory is able to create.
        """
        if locator is None:
            raise Exception("Locator cannot be null")

        # Iterate from the latest factories
        for factory in reversed(self.__factories):
            _locator = factory.can_create(locator)
            if not (_locator is None):
                return _locator

        return None

    def create(self, locator: Any) -> Any:
        """
        Creates a component identified by given locator.

        :param locator: a locator to identify component to be created.

        :return: the created component.
        """
        if locator is None:
            raise Exception("Locator cannot be null")

        # Iterate from the latest factories
        for factory in reversed(self.__factories):
            if factory.can_create(locator):
                return factory.create(locator)

        raise CreateException(None, "Cannot find factory for component " + locator)



# -*- coding: utf-8 -*-
"""
    pip_services4_components.build.IFactory
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Interface for factory components.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Any


class IFactory:
    """
    Interface for component factories.

    Factories use locators to identify components to be created.

    The locators are similar to those used to locate components in references.
    They can be of any type like strings or integers. However Pip.Services toolkit
    most often uses Descriptor objects as component locators.
    """

    def can_create(self, locator: Any) -> Any:
        """
        Checks if this factory is able to create component by given locator.

        This method searches for all registered components and returns
        a locator for component it is able to create that matches the given locator.
        If the factory is not able to create a requested component is returns null.

        :param locator: a locator to identify component to be created.

        :return: a locator for a component that the factory is able to create.
        """
        raise NotImplementedError('Method from interface definition')

    def create(self, locator: Any) -> Any:
        """
        Creates a component identified by given locator.

        :param locator: a locator to identify component to be created.

        :return: the created component.
        """
        raise NotImplementedError('Method from interface definition')



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.exec.Executor
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Executor component implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Any, List, Optional

from pip_services4_components.context.IContext import IContext

from .IExecutable import IExecutable
from .Parameters import Parameters


class Executor:
    """
    Helper class that executes components.
    """

    @staticmethod
    def execute_one(context: Optional[IContext], component: Any, args: Parameters):
        """
        Executes specific component.
        To be executed components must implement :class:`IExecutable <pip_services4_commons.exec.IExecutable.IExecutable>` interface.
        If they don't the call to this method has no effect.

        :param context: (optional) transaction id to trace execution through call chain.

        :param component: the component that is to be executed.

        :param args: execution arguments.

        :return: execution result
        """
        if isinstance(component, IExecutable):
            return component.execute(context, args)

        return None

    @staticmethod
    def execute(context: Optional[str], components: List[Any], args: Parameters = None):
        """
        Executes multiple components.

        To be executed components must implement :class:`IExecutable <pip_services4_commons.exec.IExecutable.IExecutable>` interface.
        If they don't the call to this method has no effect.

        :param context: (optional) transaction id to trace execution through call chain.

        :param components: a list of components that are to be executed.

        :param args: execution arguments.

        :return: execution result
        """
        results = []

        if components is None:
            return

        args = args if not (args is None) else Parameters()
        for component in components:
            result = Executor.execute_one(context, component, args)
            results.append(result)

        return results



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.refer.DependencyResolver
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Dependency resolver component implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Any, List, Optional

from pip_services4_commons.convert import StringConverter

from ..config import ConfigParams
from ..refer.IReferences import IReferences
from .IReferenceable import IReferenceable
from ..config.IReconfigurable import IReconfigurable
from ..refer.Descriptor import Descriptor
from ..refer.ReferenceException import ReferenceException


class DependencyResolver(IReconfigurable, IReferenceable):
    """
    Helper class for resolving component dependencies.
    The resolver is configured to resolve named dependencies by specific locator.
    During deployment the dependency locator can be changed.

    This mechanism can be used to clarify specific dependency among several alternatives.
    Typically components are configured to retrieve the first dependency that matches
    logical group, type and version. But if container contains more than one instance
    and resolution has to be specific about those instances, they can be given a unique
    name and dependency resolvers can be reconfigured to retrieve dependencies by their name.

    ### Configuration parameters ###
    dependencies:
        - [dependency name 1]: Dependency 1 locator (descriptor)
        - ...
        - [dependency name N]: Dependency N locator (descriptor)
    References:
        - References must match configured dependencies.

    Example:

    .. code-block:: python

        class MyComponent(IConfigurable, IReferenceable):
            _dependencyResolver = None
            _persistence = None

            def __init__(self):
                self._dependencyResolver.put("persistence", new Descriptor("mygroup", "persistence", "*", "*", "1.0"))

            def configure(self, config):
                self._dependencyResolver.configure(config)

            def set_references(self, references):
                self._dependencyResolver.setReferences(references)
                self._persistence = self._dependencyResolver.get_one_required("persistence")

            component = MyComponent()
            component.configure(ConfigParams.from_tuples(
            "dependencies.persistence", "mygroup:persistence:*:persistence2:1.0"))

            component.set_references(References.from_tuples(Descriptor("mygroup","persistence","*","persistence1","1.0"),
            MyPersistence(),
            Descriptor("mygroup","persistence","*","persistence2","1.0"), MyPersistence()
            # This dependency shall be set))
    """

    def __init__(self, config: ConfigParams = None, references: IReferences = None):
        """
        Creates a new instance of the dependency resolver.

        :param config: (optional) default configuration where key is dependency name and args is locator (descriptor)

        :param references: (optional) default component references
        """
        self.__dependencies: Any = {}
        self.__references: IReferences = None

        if not (config is None):
            self.configure(config)
        if not (references is None):
            self.set_references(references)

    def configure(self, config: ConfigParams):
        """
        Configures the component with specified parameters.

        :param config: configuration parameters to set.
        """
        dependencies = config.get_section("dependencies")
        names = dependencies.get_keys()
        for name in names:
            locator = dependencies.get(name)
            if locator is None:
                continue

            try:
                descriptor = Descriptor.from_string(locator)
                if not (descriptor is None):
                    self.__dependencies[name] = descriptor
                else:
                    self.__dependencies[name] = locator
            except Exception as ex:
                self.__dependencies[name] = locator

    def set_references(self, references: IReferences):
        """
        Sets the component references. References must match configured dependencies.

        :param references: references to set.
        """
        self.__references = references

    def put(self, name: str, locator: Any):
        """
        Adds a new dependency into this resolver.

        :param name: the dependency's name.

        :param locator: the locator to find the dependency by.
        """
        self.__dependencies[name] = locator

    def __locate(self, name: str) -> Any:
        """
        Gets a dependency locator by its name.

        :param name: the name of the dependency to locate.
        :return: the dependency locator or null if locator was not configured.
        """
        if name is None:
            raise Exception("Dependency name cannot be null")
        if self.__references is None:
            raise Exception("References shall be set")

        return self.__dependencies.get(name)

    def get_optional(self, name: str) -> List[Any]:
        """
        Gets all optional dependencies by their name.

        :param name: the dependency name to locate.

        :return: a list with found dependencies or empty list of no dependencies was found.
        """
        locator = self.__locate(name)
        return self.__references.get_optional(locator) if not (locator is None) else None

    def get_required(self, name: str) -> List[Any]:
        """
        Gets all __required dependencies by their name.
        At least one dependency must be present. If no dependencies was found it throws a :class:`ReferenceException <pip_services4_compoents.refer.ReferenceException.ReferenceException>`

        :param name: the dependency name to locate.

        :return: a list with found dependencies.
        """
        locator = self.__locate(name)
        if locator is None:
            raise ReferenceException(None, name)

        return self.__references.get_required(locator)

    def get_one_optional(self, name: str) -> Any:
        """
        Gets one optional dependency by its name.

        :param name: the dependency name to locate.

        :return: a dependency reference or null of the dependency was not found
        """
        locator = self.__locate(name)
        return self.__references.get_one_optional(locator) if not (locator is None) else None

    def get_one_required(self, name: str) -> Any:
        """
        Gets one __required dependency by its name.
        At least one dependency must present. If the dependency was found it throws a :class:`ReferenceException <pip_services4_compoents.refer.ReferenceException.ReferenceException>`

        :param name: the dependency name to locate.

        :return: a dependency reference
        """
        locator = self.__locate(name)
        if locator is None:
            raise ReferenceException(None, name)

        return self.__references.get_one_required(locator)

    def find(self, name: str, required: bool) -> Optional[List[Any]]:
        """
        Finds all matching dependencies by their name.

        :param name: the dependency name to locate.

        :param required: true to raise an error when no dependencies are found.

        :return: a list of found dependencies
        """
        if name is None:
            raise Exception("Name cannot be null")

        locator = self.__locate(name)
        if locator is None:
            if required:
                raise ReferenceException(None, name)
            return None

        return self.__references.find(locator, required)

    @staticmethod
    def from_tuples(*tuples: Any) -> 'DependencyResolver':
        """
        Creates a new DependencyResolver from a list of key-args pairs called tuples
        where key is dependency name and args the depedency locator (descriptor).

        :param tuples: a list of values where odd elements are dependency name
        and the following even elements are dependency locator (descriptor)

        :return: a newly created DependencyResolver.
        """
        result = DependencyResolver()
        if tuples is None or len(tuples) == 0:
            return result

        index = 0
        while index < len(tuples):
            if index + 1 >= len(tuples):
                break

            name = StringConverter.to_string(tuples[index])
            locator = tuples[index + 1]

            result.put(name, locator)
            index = index + 2

        return result



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.exec.IOpenable
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Interface for openable components
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Optional

from pip_services4_components.context.IContext import IContext

from .IClosable import IClosable


class IOpenable(IClosable):
    """
    Interface for components that require explicit opening and closing.
    For components that perform opening on demand consider using :class:`IClosable <pip_services4_components.run.IClosable.IClosable>` interface instead.

    .. code-block:: python
        class MyPersistence(IOpenable):
            _client = None
            ...
            def is_open(self):
                return self._client is not None

            def open(context):
                if self.is_open()
                    return
                ...

            def close(self, context):
                if self._client is not None:
                    self._client.close()
                    self._client = None
            ...
    """

    def is_open(self) -> bool:
        """
        Checks if the component is opened.

        :return: true if the component has been opened and false otherwise.
        """
        raise NotImplementedError('Method from interface definition')

    def open(self, context: Optional[IContext]):
        """
        Opens the component.

        :param context: (optional) transaction id to trace execution through call chain.
        """
        raise NotImplementedError('Method from interface definition')



from pip_services4_components.context import IContext


class ContextResolver:
    """
    Context resolver that processes context and extracts values from there.
    See :class:`IContext <pip_services4_components.context.IContext.IContext>`
    """

    @staticmethod
    def get_trace_id(context: IContext) -> str:
        """
        Extracts trace id from execution context.

        :param context: execution context to trace execution through call chain.
        :return: a trace id or <code>null</code> if it is not defined.

        See :class:`IContext <pip_services4_components.context.IContext.IContext>`
        """
        if not context:
            return ""
        trace_id = context.get('trace_id') or context.get('traceId')
        return trace_id or ""

    @staticmethod
    def get_client(context: IContext) -> str:
        """
        Extracts client name from execution context.

        :param context: execution context to trace execution through call chain.
        :return: a client name or <code>null</code> if it is not defined.

        See :class:`IContext <pip_services4_components.context.IContext.IContext>`

        """
        if not context:
            return ""
        trace_id = context.get('client')
        return trace_id or ""

    @staticmethod
    def get_user(context: IContext) -> str:
        """
        Extracts user name (identifier) from execution context.

        :param context: execution context to trace execution through call chain.
        :return: a user reference or <code>null</code> if it is not defined.
        """
        if not context:
            return ""
        trace_id = context.get('user')
        return trace_id or ""



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.exec.Parameters
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Parameters component implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Any

from pip_services4_commons.convert import JsonConverter
from pip_services4_commons.data import AnyValueMap
from pip_services4_commons.reflect import RecursiveObjectReader, RecursiveObjectWriter, ObjectWriter

from ..config import ConfigParams


class Parameters(AnyValueMap):
    """
    Contains map with execution parameters.

    In general, this map may contain non-serializable values.
    And in contrast with obj maps, its getters and setters
    support dot notation and able to access properties in the entire object graph.

    This class is often use to pass execution and notification
    arguments, and parameterize classes before execution.
    """

    def __init__(self, map: Any = None):
        """
        Creates a new instance of the map and assigns its args.

        :param map:(optional) values to initialize this map.
        """
        super(Parameters, self).__init__(map)

    def get(self, key: str) -> Any:
        """
        Gets a map element specified by its key.

        The key can be defined using dot notation
        and allows to recursively access elements of elements.

        :param key: a key of the element to get.

        :return: the args of the map element.
        """
        if key is None or key == '':
            return None
        elif key.find('.') > 0:
            return RecursiveObjectReader.get_property(self, key)
        else:
            return super(Parameters, self).get(key)

    def put(self, key: str, value: Any) -> Any:
        """
        Puts a new args into map element specified by its key.

        The key can be defined using dot notation
        and allows to recursively access elements of elements.

        :param key: a key of the element to put.

        :param value: a new args for map element.
        """
        if key is None or key == '':
            return None
        elif key.find('.') > 0:
            RecursiveObjectWriter.set_property(self, key, value)
            return value
        else:
            self[key] = value
            return value

    def get_as_nullable_parameters(self, key: str) -> 'Parameters':
        """
        Converts map element into an Parameters or returns null if conversion is not possible.

        :param key: a key of element to get.

        :return: Parameters args of the element or null if conversion is not supported.
        """
        value = self.get_as_nullable_map(key)
        return Parameters(value) if not (value is None) else None

    def get_as_parameters(self, key: str) -> 'Parameters':
        """
        Converts map element into an Parameters or returns empty Parameters if conversion is not possible.

        :param key: a key of element to get.

        :return: Parameters args of the element or empty Parameters if conversion is not supported.
        """
        value = self.get_as_map(key)
        return Parameters(value)

    def get_as_parameters_with_default(self, key: str, default_value: 'Parameters') -> 'Parameters':
        """
        Converts map element into an Parameters or returns default args if conversion is not possible.

        :param key: a key of element to get.

        :param default_value: the default args

        :return: Parameters args of the element or default args if conversion is not supported.
        """
        result = self.get_as_nullable_parameters(key)
        return result if not (result is None) else default_value

    def contains_key(self, key: str) -> bool:
        """
        Checks if this map contains an element with specified key.

        The key can be defined using dot notation
        and allows to recursively access elements of elements.

        :param key: a key to be checked

        :return: true if this map contains the key or false otherwise.
        """
        return RecursiveObjectReader.has_property(self, key)

    def override(self, parameters: 'Parameters', recursive: bool = False) -> 'Parameters':
        """
        Overrides parameters with new values from specified Parameters and returns a new Parameters object.

        :param parameters: Parameters with parameters to override the current values.

        :param recursive: (optional) true to perform deep copy, and false for shallow copy. Default: false

        :return: a new Parameters object.
        """
        result = Parameters()

        if recursive:
            RecursiveObjectWriter.copy_properties(result, self)
            RecursiveObjectWriter.copy_properties(result, parameters)
        else:
            ObjectWriter.set_properties(result, self)
            ObjectWriter.set_properties(result, parameters)

        return result

    def set_defaults(self, default_values: 'Parameters', recursive: bool = False) -> 'Parameters':
        """
        Set default values from specified Parameters and returns a new Parameters object.

        :param default_values: Parameters with default parameter values.

        :param recursive: (optional) true to perform deep copy, and false for shallow copy. Default: false

        :return: a new Parameters object.
        """
        result = Parameters()

        if recursive:
            RecursiveObjectWriter.copy_properties(result, default_values)
            RecursiveObjectWriter.copy_properties(result, self)
        else:
            ObjectWriter.set_properties(result, default_values)
            ObjectWriter.set_properties(result, self)

        return result

    def assign_to(self, value: Any):
        """
        Assigns (copies over) properties from the specified args to this map.

        :param value: args whose properties shall be copied over.
        """
        if value is None or len(self) == 0:
            return

        RecursiveObjectWriter.copy_properties(value, self)

    def pick(self, *props: str) -> 'Parameters':
        """
        Picks select parameters from this Parameters and returns them as a new Parameters object.

        :param props: keys to be picked and copied over to new Parameters.

        :return: a new Parameters object.
        """
        result = Parameters()
        for prop in props:
            if self.contains_key(prop):
                result.put(prop, self.get(prop))
        return result

    def omit(self, *props: str) -> 'Parameters':
        """
        Omits selected parameters from this Parameters and returns the rest as a new Parameters object.

        :param props: keys to be omitted from copying over to new Parameters.

        :return: a new Parameters object.
        """
        result = Parameters(self)
        for prop in props:
            del result[prop]
        return result

    def to_json(self) -> str:
        """
        Converts this map to JSON object.

        :return: a JSON representation of this map.
        """
        return JsonConverter.to_json(self)

    @staticmethod
    def from_value(value: Any) -> 'Parameters':
        """
        Creates a new Parameters object filled with key-args pairs from specified object.

        :param value: an object with key-args pairs used to initialize a new Parameters.

        :return: a new Parameters object.
        """
        map = value if isinstance(value, dict) else RecursiveObjectReader.get_properties(value)
        return Parameters(map)

    @staticmethod
    def from_tuples(*tuples: Any) -> 'Parameters':
        """
        Creates a new Parameters object filled with provided key-args pairs called tuples.
        Tuples parameters contain a sequence of key1, value1, key2, value2, ... pairs.

        :param tuples: the tuples to fill a new Parameters object.

        :return: a new Parameters object.
        """
        map = AnyValueMap.from_tuples_array(tuples)
        return Parameters(map)

    @staticmethod
    def merge_params(*parameters: 'Parameters') -> 'Parameters':
        """
        Merges two or more Parameters into one. The following Parameters override previously defined parameters.

        :param parameters: a list of Parameters objects to be merged.

        :return: a new Parameters object.
        """
        map = AnyValueMap.from_tuples(*parameters)
        return Parameters(map)

    @staticmethod
    def from_json(json: str) -> 'Parameters':
        """
        Creates new Parameters from JSON object.

        :param json: a JSON string containing parameters.

        :return: a new Parameters object.
        """
        map = JsonConverter.to_nullable_map(json)
        return Parameters(map)

    @staticmethod
    def from_config(config: ConfigParams) -> 'Parameters':
        """
        Creates new Parameters from ConfigMap object.

        :param config: a ConfigParams that contain parameters.

        :return: a new Parameters object.
        """
        result = Parameters()

        if config is None or len(config) == 0:
            return result

        for (key, value) in config.items():
            result.put(key, value)

        return result



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.config.IReconfigurable
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Interface for components that can be reconfigured when configuration changes
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from .IConfigurable import IConfigurable

class IReconfigurable(IConfigurable):
    """
    An interface to set configuration parameters to an object.
    It is similar to :class:`IConfigurable <pip_services4_components.config.IConfigurable.IConfigurable>` interface, but emphasises the fact
    that :func:`configure()` method can be called more than once to change object configuration in runtime.
    """
    pass


# -*- coding: utf-8 -*-
"""
    pip_services4_components.context.DefaultInfoFactory
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Default context factory implementation

    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

from .ContextInfo import ContextInfo
from ..build.Factory import Factory
from ..refer.Descriptor import Descriptor


class DefaultContextFactory(Factory):
    """
    Creates information components by their descriptors.

    See :class:`IFactory <pip_services4_components.build.IFactory.IFactory>`,
    :class:`ContextInfo <pip_services4_components.context.ContextInfo.ContextInfo>`
    """

    ContextInfoDescriptor = Descriptor("pip-services", "context-info", "default", "*", "1.0")
    ContainerInfoDescriptor = Descriptor("pip-services", "container-info", "default", "*", "1.0")

    def __init__(self):
        """
        Create a new instance of the factory.
        """
        super().__init__()
        self.register_as_type(DefaultContextFactory.ContextInfoDescriptor, ContextInfo)
        self.register_as_type(DefaultContextFactory.ContainerInfoDescriptor, ContextInfo)



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.refer.References
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Referencescomponent implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

import threading
from typing import List, Any, Sequence

from .IReferences import IReferences
from .Reference import Reference
from .ReferenceException import ReferenceException


class References(IReferences):
    """
    The most basic implementation of :class:`IReferences <pip_services4_components.refer.IReferences.IReferences>` to store and locate component references.

    Example:

    .. code-block:: python

        class MyController(IReferenceable):
            _persistence = None

            def set_references(self, references):
                self._persistence = references.getOneRequired(Descriptor("mygroup", "persistence", "*", "*", "1.0"))

        persistence = MyMongoDbPersistence()

        references = References.from_tuples(
        Descriptor("mygroup", "persistence", "mongodb", "default", "1.0"), persistence,
        Descriptor("mygroup", "controller", "default", "default", "1.0"), controller)

        controller.set_references(references)


    """

    __lock = None

    def __init__(self, tuples: Sequence[Any] = None):
        """
        Creates a new instance of references and initializes it with references.

        :param tuples: (optional) a list of values where odd elements are locators
        and the following even elements are component references
        """
        self._references: List[Reference] = []
        self.__lock = threading.Lock()

        if not (tuples is None):
            index = 0
            while index < len(tuples):
                if index + 1 >= len(tuples):
                    break
                self.put(tuples[index], tuples[index + 1])
                index = index + 2

    def put(self, locator: Any = None, component: Any = None):
        """
        Puts a new reference into this reference map.

        :param locator: a component reference to be added.

        :param component: a locator to find the reference by.
        """
        if component is None:
            raise Exception("Component cannot be null")

        self.__lock.acquire()
        try:
            self._references.append(Reference(locator, component))
        finally:
            self.__lock.release()

    def remove(self, locator: Any) -> Any:
        """
        Removes a previously added reference that matches specified locator.
        If many references match the locator, it removes only the first one.
        When all references shall be removed, use :func:`remove_all` method instead.

        :param locator: a locator to remove reference

        :return: the removed component reference.
        """
        if locator is None:
            return None

        self.__lock.acquire()
        try:
            for reference in reversed(self._references):
                if reference.match(locator):
                    self._references.remove(reference)
                    return reference.get_component()
        finally:
            self.__lock.release()

        return None

    def remove_all(self, locator: Any) -> List[Any]:
        """
        Removes all component references that match the specified locator.

        :param locator: a locator to remove reference by.

        :return: a list, containing all removed references.
        """
        components = []

        if locator is None:
            return components

        self.__lock.acquire()
        try:
            for reference in reversed(self._references):
                if reference.match(locator):
                    self._references.remove(reference)
                    components.append(reference.get_component())
        finally:
            self.__lock.release()

        return components

    def get_all_locators(self) -> List[Any]:
        """
        Gets locators for all registered component references in this reference map.

        :return: a list with component locators.
        """
        locators = []

        self.__lock.acquire()
        try:
            for reference in self._references:
                locators.append(reference.get_locator())
        finally:
            self.__lock.release()

        return locators

    def get_all(self) -> List[Any]:
        """
        Gets all component references registered in this reference map.

        :return: a list with component references.
        """
        components = []

        self.__lock.acquire()
        try:
            for reference in self._references:
                components.append(reference.get_component())
        finally:
            self.__lock.release()

        return components

    def get_optional(self, locator: Any) -> List[Any]:
        """
        Gets all component references that match specified locator.

        :param locator: the locator to find references by.

        :return: a list with matching component references or empty list if nothing was found.
        """
        try:
            return self.find(locator, False)
        except Exception as ex:
            return []

    def get_required(self, locator: Any) -> List[Any]:
        """
        Gets all component references that match specified locator.
        At least one component reference must be present. If it doesn't the method throws an error.

        :param locator: the locator to find references by.

        :return: a list with matching component references.

        :raises: a :class:`ReferenceException <pip_services4_compoents.refer.ReferenceException.ReferenceException>` when no references found.
        """
        return self.find(locator, True)

    def get_one_optional(self, locator: Any) -> Any:
        """
        Gets an optional component reference that matches specified locator.

        :param locator: the locator to find references by.

        :return: a matching component reference or null if nothing was found.
        """
        try:
            components = self.find(locator, False)
            return components[0] if len(components) > 0 else None
        except Exception as ex:
            return None

    def get_one_required(self, locator: Any) -> Any:
        """
         Gets a __required component reference that matches specified locator.

         :param locator: the locator to find a reference by.

         :return: a matching component reference.

         :raises: a :class:`ReferenceException <pip_services4_compoents.refer.ReferenceException.ReferenceException>` when no references found.
         """
        components = self.find(locator, True)
        return components[0] if len(components) > 0 else None

    def find(self, locator: Any, required: bool) -> List[Any]:
        """
        Gets all component references that match specified locator.

        :param locator: the locator to find a reference by.

        :param required: forces to raise an error if no reference is found.

        :return: a list with matching component references.

        :raises: a :class:`ReferenceException <pip_services4_compoents.refer.ReferenceException.ReferenceException>` when __required is set to true but no references found.
        """
        if locator is None:
            raise Exception("Locator cannot be null")

        components = []

        self.__lock.acquire()
        try:
            index = len(self._references) - 1

            while index >= 0:
                reference = self._references[index]
                if reference.match(locator):
                    component = reference.get_component()
                    components.append(component)
                index = index - 1

            if len(components) == 0 and required:
                raise ReferenceException(None, locator)
        finally:
            self.__lock.release()

        return components

    @staticmethod
    def from_tuples(*tuples: Any) -> 'References':
        """
        Creates a new References from a list of key-args pairs called tuples.

        :param tuples: a list of values where odd elements are locators
                      and the following even elements are component references

        :return: a newly created References.
        """
        return References(tuples)



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.refer.IReferenceable
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Interface for referenceable components.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from abc import ABC

from ..refer import IReferences


class IReferenceable(ABC):
    """
    Interface for components that depends on obj components.
    If component requires explicit notification to unset references
    it shall additionally implement :class:`IUnreferenceable <pip_services4_compoents.refer.IUnreferenceable.IUnreferenceable>` interface.

    Example:

    .. code-block:: python
    
        class MyController(IReferenceable):
            _persistence = None

            def set_references(self, references):
                self._persistence = references.get_one_required(Descriptor("mygroup", "persistence", "*", "*", "1.0"))
    """

    def set_references(self, references: IReferences):
        """
        Sets references to dependent components.

        :param references: references to locate the component dependencies.
        """
        raise NotImplementedError('Method from interface definition')



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.config.IConfigurable
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Interface for components that require configuration
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from abc import ABC

from ..config.ConfigParams import ConfigParams


class IConfigurable(ABC):
    """
     An interface to set configuration parameters to an object.

     It can be added to any existing class by implementing a single :func:`configure()` method.

     If you need to emphasis the fact that :func:`configure()` method can be called multiple times
     to change object configuration in runtime, use :class:`IReconfigurable <pip_services4_components.config.IReconfigurable.IReconfigurable>` interface instead.

     Example:

    .. code-block:: python

         class MyClass(IConfigurable):
            _myParam = "default args"

         def configure(self, config):
            self._myParam = config.get_as_string_with_default("options.param", myParam)
    """

    def configure(self, config: ConfigParams):
        """
        Configures object by passing configuration parameters.

        :param config: configuration parameters to be set.
        """
        raise NotImplementedError('Method from interface definition')



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.exec.IClosable
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Interface for closable components
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from abc import ABC
from typing import Optional

from pip_services4_components.context.IContext import IContext


class IClosable(ABC):
    """
    Interface for components that require explicit closure.

    For components that require opening as well as closing use :class:`IOpenable <pip_services4_components.run.IOpenable.IOpenable>` interface instead.

    .. code-block:: python
        class MyConnector(ICloseable):
            _client = None

            ... # The _client can be lazy created

            def close(self, context):
                if self._client is not None:
                    self._client.close()
                    self._client = null


    """

    def close(self, context: Optional[IContext]):
        """
        Closes component and frees used resources.

        :param context: (optional) transaction id to trace execution through call chain.
        """
        raise NotImplementedError('Method from interface definition')



# -*- coding: utf-8 -*-

class ClassTest:

    def __init__(self, value1, value2):
        self.value1 = value1
        self._value2 = value2

    @property
    def value2(self):
        return self._value2

    @value2.setter
    def value2(self, value):
        self._value2 = value



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.exec.Notifier
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Notifier component implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Any, List, Optional

from pip_services4_components.context.IContext import IContext

from .INotifiable import INotifiable
from .Parameters import Parameters


class Notifier:
    """
    Helper class that notifies components.
    """

    @staticmethod
    def notify_one(context: Optional[IContext], component: Any, args: Parameters):
        """
        Notifies specific component.
        To be notiied components must implement :class:`INotifiable <pip_services4_components.exec.INotifiable.INotifiable>` interface.
        If they don't the call to this method has no effect.

        :param context: (optional) transaction id to trace execution through call chain.

        :param component: the component that is to be notified.

        :param args: notifiation arguments.
        """
        if component is None:
            return

        if isinstance(component, INotifiable):
            component.notify(context, args)

    @staticmethod
    def notify(context: Optional[IContext], components: List[Any], args: Parameters = None):
        """
        Notifies multiple components.

        To be notified components must implement :class:`INotifiable <pip_services4_components.exec.INotifiable.INotifiable>` interface.
        If they don't the call to this method has no effect.

        :param context: (optional) transaction id to trace execution through call chain.

        :param components: a list of components that are to be notified.

        :param args: notification arguments.
        """
        if components is None:
            return

        args = args if not (args is None) else Parameters()
        for component in components:
            Notifier.notify_one(context, component, args)



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.exec.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Contains design patterns for the standard lifecycle of objects (opened,
    closed, openable, closable, runnable). Helper classes for lifecycle provisioning.

    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = [
    'Parameters', 'IParameterized', 'FixedRateTimer',
    'Cleaner',
    'IExecutable', 'Executor',
    'INotifiable', 'Notifier'
]

from pip_services4_components.run.Cleaner import Cleaner
from .Executor import Executor
from .FixedRateTimer import FixedRateTimer
from .IExecutable import IExecutable
from .INotifiable import INotifiable
from .IParameterized import IParameterized
from .Notifier import Notifier
from .Parameters import Parameters



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.exec.Opener
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Opener component implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Any, List, Optional

from pip_services4_components.context.IContext import IContext

from .IOpenable import IOpenable


class Opener:
    """
    Helper class that opens components.
    """

    @staticmethod
    def is_opened_one(component: Any) -> bool:
        """
        Checks if specified component is opened.

        To be checked components must implement :class:`IOpenable <pip_services4_components.run.IOpenable.IOpenable>` interface.
        If they don't the call to this method returns true.

        :param component: the component that is to be checked.

        :return: true if component is opened and false otherwise.
        """
        if isinstance(component, IOpenable):
            return component.is_open()

        return True

    @staticmethod
    def is_opened(components: List[Any]) -> bool:
        """
        Checks if all components are opened.

        To be checked components must implement :class:`IOpenable <pip_services4_components.run.IOpenable.IOpenable>` interface.
        If they don't the call to this method returns true.

        :param components: a list of components that are to be checked.

        :return: true if all components are opened and false if at least one component is closed.
        """
        if components is None:
            return True

        result = True
        for component in components:
            result = result and Opener.is_opened_one(component)

        return result

    @staticmethod
    def open_one(context: Optional[IContext], component: Any):
        """
        Opens specific component.

        To be opened components must implement :class:`IOpenable <pip_services4_components.run.IOpenable.IOpenable>` interface.
        If they don't the call to this method has no effect.

        :param context: (optional) transaction id to trace execution through call chain.

        :param component: the component that is to be opened.
        """
        if isinstance(component, IOpenable):
            component.open(context)

    @staticmethod
    def open(context: Optional[IContext], components: List[Any]):
        """
        Opens multiple components.

        To be opened components must implement :class:`IOpenable <pip_services4_components.run.IOpenable.IOpenable>` interface.
        If they don't the call to this method has no effect.

        :param context: (optional) transaction id to trace execution through call chain.

        :param components: the list of components that are to be closed.
        """
        if components is None:
            return

        for component in components:
            Opener.open_one(context, component)



# -*- coding: utf-8 -*-
"""
    pip_services4_components.context.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Contains a simple object that defines the context of execution. For various
    logging functions we need to know what source we are logging from – what is
    the processes name, what the process is/does.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = ['ContextInfo', 'DefaultContextFactory', 'IContext', 'Context', 'ContextResolver']

from .ContextInfo import ContextInfo
from .DefaultContextFactory import DefaultContextFactory
from .IContext import IContext
from .Context import Context
from .ContextResolver import ContextResolver


# -*- coding: utf-8 -*-
"""
    pip_services4_components.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Components module initialization
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = []




# -*- coding: utf-8 -*-
"""
    pip_services4_commons.config.ConfigParams
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Config params implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Any, List

from pip_services4_commons.data import StringValueMap
from pip_services4_commons.reflect import RecursiveObjectReader


class ConfigParams(StringValueMap):
    """
    Contains a key-args map with configuration parameters.
    All values stored as strings and can be serialized as JSON or string forms.
    When retrieved the values can be automatically converted on read using GetAsXXX methods.

    The keys are case-sensitive, so it is recommended to use consistent C-style as: **"my_param"**

    Configuration parameters can be broken into sections and subsections using dot notation as:
    **"section1.subsection1.param1"**. Using GetSection method all parameters from specified section
    can be extracted from a ConfigMap.

    The ConfigParams supports serialization from/to plain strings as:
    **"key1=123;key2=ABC;key3=2016-09-16T00:00:00.00Z"**

    ConfigParams are used to pass configurations to :class:`IConfigurable <pip_services4_components.config.IConfigurable.IConfigurable>` objects.
    They also serve as a basis for more concrete configurations such as :class:`ConnectionParams <pip_services4_config.connect.ConnectionParams.ConnectionParams>`
    or :class:`CredentialParams <pip_services4_config.auth.CredentialParams.CredentialParams>` (in the Pip.Services components package).

    Example:

    .. code-block:: python

        config = ConfigParams.fromTuples("section1.key1", "AAA",
                                         "section1.key2", 123,
                                         "section2.key1", true)
        config.get_as_string("section1.key1")  # Result: AAA
        config.get_as_integer("section1.key1") # Result: 0

        section1 = config.get_section("section1")
        section1.__str__() # Result: key1=AAA;key2=123
    """

    def __init__(self, values: Any = None):
        """
        Creates a new :class:`ConfigParams` and fills it with values.

        :param values: (optional) an object to be converted into key-args pairs to initialize this config map.
        """
        super().__init__(values)

    def get_section_names(self) -> List[str]:
        """
        Gets a list with all 1st level section names.

        :return: a list of section names stored in this ConfigMap.
        """
        sections = []

        for (key, value) in self.items():
            pos = key.find('.')
            if pos > 0:
                key = key[0: pos]

            # Perform case sensitive search
            found = False
            for section in sections:
                if section == key:
                    found = True
                    break

            if not found:
                sections.append(key)

        return sections

    def get_section(self, section: str) -> 'ConfigParams':
        """
        Gets parameters from specific section stored in this ConfigMap.
        The section name is removed from parameter keys.

        :param section: name of the section to retrieve configuration parameters from.

        :return: all configuration parameters that belong to the section named 'section'.
        """
        result = ConfigParams()
        prefix = section + "."

        for (key, value) in self.items():
            # Prevents error on the next line
            if len(key) < len(prefix):
                continue

            # Perform case sensitive match
            key_prefix = key[: len(prefix)]
            if key_prefix == prefix:
                key = key[len(prefix):]
                result[key] = value

        return result

    def _is_shadow_name(self, name):
        return name is None or len(name) == 0 or name[0] == "#" or name[0] == "!"

    def add_section(self, section: str, section_params: 'ConfigParams'):
        """
        Adds parameters into this ConfigParams under specified section.
        Keys for the new parameters are appended with section dot prefix.

        :param section: name of the section where add new parameters

        :param section_params: new parameters to be added.
        """
        if section is None:
            raise Exception("Section name cannot be null")

        section = "" if self._is_shadow_name(section) else section

        if section_params is None or len(section_params) == 0:
            return

        for (key, value) in section_params.items():
            key = "" if self._is_shadow_name(key) else key

            if len(key) > 0 and len(section) > 0:
                key = section + "." + key
            elif len(key) == 0:
                key = section

            self[key] = value

    def override(self, config_params: 'ConfigParams') -> 'ConfigParams':
        """
        Overrides parameters with new values from specified ConfigParams and returns a new ConfigParams object.

        :param config_params: ConfigMap with parameters to override the current values.

        :return: a new ConfigParams object.
        """
        map = StringValueMap.from_maps(self, config_params)
        return ConfigParams(map)

    def set_defaults(self, default_config_params: 'ConfigParams') -> 'ConfigParams':
        """
        Set default values from specified ConfigParams and returns a new ConfigParams object.

        :param default_config_params: ConfigMap with default parameter values.

        :return: a new ConfigParams object.
        """
        map = StringValueMap.from_maps(default_config_params, self)
        return ConfigParams(map)

    @staticmethod
    def from_value(value: Any) -> 'ConfigParams':
        """
        Creates a new ConfigParams object filled with key-args pairs from specified object.

        :param value: an object with key-args pairs used to initialize a new ConfigParams.

        :return: a new ConfigParams object.
        """
        map = RecursiveObjectReader.get_properties(value)
        return ConfigParams(map)

    @staticmethod
    def from_tuples(*tuples: Any) -> 'ConfigParams':
        """
        Creates a new ConfigParams object filled with provided key-args pairs called tuples.
        Tuples parameters contain a sequence of key1, value1, key2, value2, ... pairs.

        :param tuples: the tuples to fill a new ConfigParams object.

        :return: a new ConfigParams object.
        """

        map = StringValueMap.from_tuples_array(tuples)
        return ConfigParams(map)

    @staticmethod
    def from_string(line: str) -> 'ConfigParams':
        """
        Creates a new ConfigParams object filled with key-args pairs serialized as a string.

        :param line: a string with serialized key-args pairs as "key1=value1;key2=value2;..."

        :return: a new ConfigParams object.
        """
        map = StringValueMap.from_string(line)
        return ConfigParams(map)

    @staticmethod
    def merge_configs(*configs: 'ConfigParams') -> 'ConfigParams':
        """
        Merges two or more ConfigParams into one. The following ConfigParams override previously defined parameters.

        :param configs: a list of ConfigParams objects to be merged.

        :return: a new ConfigParams object.
        """
        map = StringValueMap.from_maps(*configs)
        return ConfigParams(map)



# -*- coding: utf-8 -*-
"""
    pip_services4_components.build.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Contains the "factory design pattern". There are various factory types,
    which are also implemented in a portable manner.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = ['IFactory', 'CreateException', 'CompositeFactory', 'Factory']

from .CompositeFactory import CompositeFactory
from .CreateException import CreateException
from .Factory import Factory
from .IFactory import IFactory



# -*- coding: utf-8 -*-
"""
    pip_services_runtime.refer.Descriptor
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Component descriptor implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Optional, Any

from pip_services4_commons.errors import ConfigException


class Descriptor:
    """
    Locator type that most often used in PipServices toolkit.
    It locates components using several fields:
        - Group: a package or just named group of components like "pip-services"
        - Type: logical component type that defines it's contract like "persistence"
        - Kind: physical implementation type like "mongodb"
        - Name: unique component name like "default"
        - Version: version of the component contract like "1.0"

    The locator matching can be done by all or only few selected fields.
    The fields that shall be excluded from the matching must be set to **"*"** or **None**.
    That approach allows to implement many interesting scenarios. For instance:
        - Locate all loggers (match by type and version)
        - Locate persistence components for a microservice (match by group and type)
        - Locate specific component by its name (match by name)

    Example:

    .. code-block:: python
    
        locator1 = Descriptor("mygroup", "connector", "aws", "default", "1.0")
        locator2 = Descriptor.from_string("mygroup:connector:*:*:1.0")

        locator1.match(locator2);		// Result: true
        locator1.eq(locator2);		// Result: true
        locator1.exact_match(locator2);	// Result: false
    """

    def __init__(self, group: Optional[str], type: Optional[str], kind: Optional[str], name: Optional[str],
                 version: Optional[str]):
        """
        Creates a new instance of the descriptor.

        :param group: a logical component group

        :param type: a logical component type or contract

        :param kind: a component implementation type

        :param name: a unique component name

        :param version: a component implementation version
        """
        group = None if "*" == group else group
        type = None if "*" == type else type
        kind = None if "*" == kind else kind
        name = None if "*" == name else name
        version = None if "*" == version else version

        self.__group: str = group
        self.__type: str = type
        self.__kind: str = kind
        self.__name: str = name
        self.__version: str = version

    def get_group(self) -> str:
        """
         Gets the component's logical group.

        :return: the component's logical group
        """
        return self.__group

    def get_type(self) -> str:
        """
        Gets the component's logical type.

        :return: the component's logical type.
        """
        return self.__type

    def get_kind(self) -> str:
        """
        Gets the component's implementation type.

        :return: the component's implementation type.
        """
        return self.__kind

    def get_name(self) -> str:
        """
        Gets the unique component's name.

        :return: the unique component's name.
        """
        return self.__name

    def get_version(self) -> str:
        """
        Gets the component's implementation version.

        :return: the component's implementation version.
        """
        return self.__version

    def __match_field(self, field1: str, field2: str) -> bool:
        return field1 is None \
               or field2 is None \
               or field1 == field2

    def match(self, descriptor: 'Descriptor') -> bool:
        """
        Partially matches this descriptor to another descriptor.
        Fields that contain "*" or null are excluded from the match.

        :param descriptor: the descriptor to match this one against.

        :return: true if descriptors match and false otherwise
        """
        return self.__match_field(self.__group, descriptor.get_group()) \
               and self.__match_field(self.__type, descriptor.get_type()) \
               and self.__match_field(self.__kind, descriptor.get_kind()) \
               and self.__match_field(self.__name, descriptor.get_name()) \
               and self.__match_field(self.__version, descriptor.get_version())

    def __exact_match_field(self, field1: str, field2: str) -> bool:
        if field1 is None and field2 is None:
            return True
        if field1 is None or field2 is None:
            return False
        return field1 == field2

    def exact_match(self, descriptor: 'Descriptor') -> bool:
        """
        Matches this descriptor to another descriptor by all fields. No exceptions are made.

        :param descriptor: the descriptor to match this one against.

        :return: true if descriptors match and false otherwise.
        """
        return self.__exact_match_field(self.__group, descriptor.get_group()) \
               and self.__exact_match_field(self.__type, descriptor.get_type()) \
               and self.__exact_match_field(self.__kind, descriptor.get_kind()) \
               and self.__exact_match_field(self.__name, descriptor.get_name()) \
               and self.__exact_match_field(self.__version, descriptor.get_version())

    def is_complete(self) -> bool:
        """
        Checks whether all descriptor fields are set.
        If descriptor has at least one "*" or null field it is considered "incomplete"

        :return: true if all descriptor fields are defined and false otherwise.
        """
        return not (self.__group is None) and not (self.__type is None) \
               and not (self.__kind is None) and not (self.__name is None) and not (self.__version is None)

    def equals(self, value: Any) -> bool:
        """
        Compares this descriptor to a args.
        If args is a Descriptor it tries to match them, otherwise the method returns false.

        :param value: the args to match against this descriptor.

        :return: true if the args is matching descriptor and false otherwise.
        """
        if isinstance(value, Descriptor):
            return self.match(value)
        return False

    def to_string(self) -> str:
        """
        Gets a string representation of the object.
        The result is a colon-separated list of descriptor fields as "mygroup:connector:aws:default:1.0"

        :return: a string representation of the object.
        """
        result = ''
        result += self.__group if not (self.__group is None) else '*'
        result += ':'
        result += self.__type if not (self.__type is None) else '*'
        result += ':'
        result += self.__kind if not (self.__kind is None) else '*'
        result += ':'
        result += self.__name if not (self.__name is None) else '*'
        result += ':'
        result += self.__version if not (self.__version is None) else '*'
        return result

    def __eq__(self, other):
        """
        Compares this descriptor to a args.
        If args is a Descriptor it tries to match them, otherwise the method returns false.

        :param other: the args to match against this descriptor.

        :return: true if the args is matching descriptor and false otherwise.
        """
        if isinstance(other, Descriptor):
            return self.match(other)
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        """
        Gets a string representation of the object.
        The result is a colon-separated list of descriptor fields as "mygroup:connector:aws:default:1.0"

        :return: a string representation of the object.
        """
        result = ''
        result += self.__group if not (self.__group is None) else '*'
        result += ':'
        result += self.__type if not (self.__type is None) else '*'
        result += ':'
        result += self.__kind if not (self.__kind is None) else '*'
        result += ':'
        result += self.__name if not (self.__name is None) else '*'
        result += ':'
        result += self.__version if not (self.__version is None) else '*'
        return result

    @staticmethod
    def from_string(value: str) -> Optional['Descriptor']:
        """
        Parses colon-separated list of descriptor fields and returns them as a Descriptor.

        :param value: colon-separated descriptor fields to initialize Descriptor.

        :return: a newly created Descriptor.
        """
        if value is None or len(value) == 0:
            return None

        tokens = value.split(":")
        if len(tokens) != 5:
            raise ConfigException(
                None, "BAD_DESCRIPTOR", "Descriptor " + str(value) + " is in wrong format"
            ).with_details("descriptor", value)

        return Descriptor(tokens[0].strip(), tokens[1].strip(), tokens[2].strip(), tokens[3].strip(), tokens[4].strip())



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.config.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Contains the implementation of the config design pattern.
    The :class:`IConfigurable <pip_services4_components.config.IConfigurable.IConfigurable>` configurable interface
    contains just one method - "configure", which takes ConfigParams as a parameter (extends StringValueMap class).
    If any object needs to be configurable, we implement this interface
    and parse the ConfigParams that the method received.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = [
    'IConfigurable', 'IReconfigurable', 'ConfigParams',
    'NameResolver', 'OptionsResolver'
]

from .ConfigParams import ConfigParams
from .IConfigurable import IConfigurable
from .IReconfigurable import IReconfigurable
from .NameResolver import NameResolver
from .OptionsResolver import OptionsResolver



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.exec.ICleanable
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Interface for cleanable components
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from abc import ABC
from typing import Optional

from pip_services4_components.context.IContext import IContext


class ICleanable(ABC):
    """
    Interface for components that should clean their state.
    Cleaning state most often is used during testing.
    But there may be situations when it can be done in production.

    .. code-block:: python
        class MyObjectWithState(ICleanable):
            _state = {}
            ...

            def clear(self, context):
                self._state = {}
    """

    def clear(self, context: Optional[IContext]):
        """
        Clears component state.

        :param context: (optional) transaction id to trace execution through call chain.
        """
        raise NotImplementedError('Method from interface definition')



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.refer.Reference
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Reference component implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Any


class Reference(object):
    """
    Contains a reference to a component and locator to find it.
    It is used by :class:`References <pip_services4_commons.refer.References.References>` to store registered component references.
    """

    def __init__(self, locator: Any, component: Any):
        """
        Create a new instance of the reference object and assigns its values.

        :param locator: a locator to find the reference.

        :param component: a reference to component.
        """
        if component is None:
            raise Exception("Component cannot be null")

        self.__locator: Any = locator
        self.__component: Any = component

    def match(self, locator: Any) -> bool:
        """
        Matches locator to this reference locator. Descriptors are matched using equal method.
        All obj locator types are matched using direct comparison.

        :param locator: the locator to match.

        :return: true if locators are matching and false it they don't.
        """
        # Locate by direct reference matching
        if self.__component == locator:
            return True
        # Locate by direct locator matching
        elif not (self.__locator is None):
            return self.__locator == locator
        else:
            return False

    def get_component(self) -> Any:
        """
        Gets the stored component reference.

        :return: the component's references.
        """
        return self.__component

    def get_locator(self) -> Any:
        """
        Gets the stored component locator.

        :return: the component's locator.
        """
        return self.__locator



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.refer.IReferences
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Interface for references components.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from abc import ABC
from typing import Any, List




class IReferences(ABC):
    """
    Interface for a map that holds component references and passes them to components
    to establish dependencies with each obj.

    Together with :class:`IReferenceable <pip_services4_commons.refer.IReferenceable.IReferenceable>` and :class:`IUnreferenceable <pip_services4_compoents.refer.IUnreferenceable.IUnreferenceable>` interfaces it implements
    a Locator pattern that is used by PipServices toolkit for Inversion of Control
    to assign external dependencies to components.

    The IReferences object is a simple map, where keys are locators
    and values are component references. It allows to add, remove and find components
    by their locators. Locators can be any values like integers, strings or component types.
    But most often PipServices toolkit uses :class:`Descriptor <pip_services4_commons.refer.Descriptor.Descriptor>` as locators that match
    by 5 fields: group, type, kind, name and version.

    Example:

    .. code-block:: python

        class MyController(IReferences):
            _persistence = None

            def set_references(self, references):
                self._persistence = references.get_one_required(Descriptor("mygroup", "persistence", "*", "*", "1.0"))

        persistence = MyMongoDbPersistence()

        references = References.from_tuples(
        Descriptor("mygroup", "persistence", "mongodb", "default", "1.0"), persistence,
        Descriptor("mygroup", "controller", "default", "default", "1.0"), controller)

        controller.set_references(references)
    """

    def put(self, locator: Any = None, reference: Any = None):
        """
        Puts a new reference into this reference map.

        :param locator: a component reference to be added.

        :param reference: a locator to find the reference by.
        """
        raise NotImplementedError('Method from interface definition')

    def remove(self, locator: Any) -> Any:
        """
        Removes a previously added reference that matches specified locator.
        If many references match the locator, it removes only the first one.
        When all references shall be removed, use :func:remove_all method instead.

        :param locator: a locator to remove reference

        :return: the removed component reference.
        """
        raise NotImplementedError('Method from interface definition')

    def remove_all(self, locator: Any) -> List[Any]:
        """
        Removes all component references that match the specified locator.

        :param locator: a locator to remove reference
        :return: a list, containing all removed references.
        """
        raise NotImplementedError('Method from interface definition')

    def get_all_locators(self) -> List[Any]:
        """
        Gets locators for all registered component references in this reference map.

        :return: a list with component locators.
        """
        raise NotImplementedError('Method from interface definition')

    def get_all(self) -> List[Any]:
        """
        Gets all component references registered in this reference map.

        :return: a list with component references.
        """
        raise NotImplementedError('Method from interface definition')

    def get_optional(self, locator: Any) -> List[Any]:
        """
        Gets all component references that match specified locator.

        :param locator: the locator to find references by.

        :return: a list with matching component references or empty list if nothing was found.
        """
        raise NotImplementedError('Method from interface definition')

    def get_required(self, locator: Any) -> List[Any]:
        """
        Gets all component references that match specified locator.
        At least one component reference must be present. If it doesn't the method throws an error.

        :param locator: the locator to find references by.

        :return: a list with matching component references.

        :raises: a :class:`ReferenceException <pip_services4_compoents.refer.ReferenceException.ReferenceException>` when no references found.
        """
        raise NotImplementedError('Method from interface definition')

    def get_one_optional(self, locator: Any) -> Any:
        """
        Gets an optional component reference that matches specified locator.

        :param locator: the locator to find references by.

        :return: a matching component reference or null if nothing was found.
        """
        raise NotImplementedError('Method from interface definition')

    def get_one_required(self, locator: Any) -> Any:
        """
        Gets a __required component reference that matches specified locator.

        :param locator: the locator to find a reference by.

        :return: a matching component reference.

        :raises: a :class:`ReferenceException <pip_services4_compoents.refer.ReferenceException.ReferenceException>` when no references found.
        """
        raise NotImplementedError('Method from interface definition')

    def get_one_before(self, reference, locator):
        raise NotImplementedError('Method from interface definition')

    def find(self, locator: Any, required: bool) -> List[Any]:
        """
        Gets all component references that match specified locator.

        :param locator: the locator to find a reference by.

        :param required: forces to raise an error if no reference is found.

        :return: a list with matching component references.

        :raises: a :class:`ReferenceException <pip_services4_compoents.refer.ReferenceException.ReferenceException>` when __required is set to true but no references found.
        """
        raise NotImplementedError('Method from interface definition')



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.connect.DefaultDiscoveryFactory
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Default discovery factory implementation

    :copyright: Conceptual Vision Consulting LLC 2015-2016, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

import datetime

from pip_services4_components.config import ConfigParams
from pip_services4_components.context.ContextInfo import ContextInfo


class TestContextInfo:
    def test_name(self):
        context_info = ContextInfo("unknown")
        assert context_info.name == "unknown"

        update = "new name"
        context_info.name = update
        assert context_info.name == "new name"

    def test_description(self):
        context_info = ContextInfo()
        assert context_info.description is None

        update = "new description"
        context_info.description = update
        assert context_info.description == "new description"

    def test_context_id(self):
        context_info = ContextInfo()
        assert context_info.context_id is not None

        update = "new context_id"
        context_info.context_id = update
        assert context_info.context_id == "new context_id"

    def test_start_time(self):
        context_info = ContextInfo()
        now = datetime.datetime.now()
        assert context_info.start_time.year == now.year
        assert context_info.start_time.month == now.month

        context_info.start_time = datetime.datetime(1975, 4, 8)
        assert context_info.start_time.year == 1975
        assert context_info.start_time.month == 4
        assert context_info.start_time.day == 8

    def test_from_config(self):
        config = ConfigParams.from_tuples("name", "new name",
                                          "description", "new description",
                                          "properties.access_key", "key",
                                          "properties.store_key", "store key")
        context_info = ContextInfo.from_config(config)
        assert context_info.name == "new name"
        assert context_info.description == "new description"



# -*- coding: utf-8 -*-
"""
    tests.exec.test_Parameters
    ~~~~~~~~~~~~~~~~~~~~~~~~~
    
    :copyright: (c) Conceptual Vision Consulting LLC 2015-2016, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from pip_services4_components.config import ConfigParams
from pip_services4_components.exec.Parameters import Parameters
from .ClassTest import ClassTest


class TestParameters:

    def test_defaults(self):
        result = Parameters.from_tuples(
            "value1", 123,
            "value2", 234
        )
        defaults = Parameters.from_tuples(
            "value2", 432,
            "value3", 345
        )
        result = result.set_defaults(defaults, False)
        assert 3 == len(result)
        assert 123 == result.get("value1")
        assert 234 == result.get("value2")
        assert 345 == result.get("value3")

    def test_override_recursive(self):
        result = Parameters.from_json(
            "{ \"value1\": 123, \"value2\": { \"value21\": 111, \"value22\": 222 } }"
        )
        defaults = Parameters.from_json(
            "{ \"value2\": { \"value22\": 777, \"value23\": 333 }, \"value3\": 345 }"
        )
        result = result.set_defaults(defaults, True)

        assert 3 == len(result)
        assert 123 == result.get("value1")
        assert 345 == result.get("value3")

        deep_result = result.get_as_map("value2")
        assert 3 == len(deep_result)
        assert 111 == deep_result.get("value21")
        assert 222 == deep_result.get("value22")
        assert 333 == deep_result.get("value23")

    def test_override_with_nulls(self):
        result = Parameters.from_json(
            "{ \"value1\": 123, \"value2\": 234 }"
        )
        result = result.override(None, True)

        assert 2 == len(result)
        assert 123 == result.get("value1")
        assert 234 == result.get("value2")

    def test_assign_to(self):
        value = ClassTest(None, None)
        new_values = Parameters.from_json(
            "{ \"value1\": 123, \"value2\": \"ABC\", \"value3\": 456 }"
        )
        new_values.assign_to(value)
        assert value.value1 is not None
        assert value.value1 == 123
        assert value.value2 is not None
        assert value.value2 == 'ABC'

    def test_get(self):
        config = Parameters.from_json(
            "{ \"value1\": 123, \"value2\": { \"value21\": 111, \"value22\": 222 } }"
        )

        value = config.get("")
        assert None == value

        value = config.get("value1")
        assert 123 == value

        value = config.get("value2")
        assert None != value

        value = config.get("value3")
        assert None == value

        value = config.get("value2.value21")
        assert 111 == value

        value = config.get("value2.value31")
        assert None == value

        value = config.get("value2.value21.value211")
        assert None == value

        value = config.get("valueA.valueB.valueC")
        assert None == value

    def test_contains(self):
        config = Parameters.from_json(
            "{ \"value1\": 123, \"value2\": { \"value21\": 111, \"value22\": 222 } }"
        )

        has = config.contains_key("")
        assert False == has

        has = config.contains_key("value1")
        assert True == has

        has = config.contains_key("value2")
        assert True == has

        has = config.contains_key("value3")
        assert False == has

        has = config.contains_key("value2.value21")
        assert True == has

        has = config.contains_key("value2.value31")
        assert False == has

        has = config.contains_key("value2.value21.value211")
        assert False == has

        has = config.contains_key("valueA.valueB.valueC")
        assert False == has

    def test_put(self):
        config = Parameters(self)

        config.put(None, 123)
        assert 0 == len(config)

        config.put("field1", 123)
        assert 1 == len(config)
        assert 123 == config.get("field1")

        config.put("field2", "ABC")
        assert 2 == len(config)
        assert "ABC" == config.get("field2")

        config.put("field2.field1", 123)
        assert "ABC" == config.get("field2")

        config.put("field3.field31", 456)
        assert 3 == len(config)
        sub_config = config.get_as_map("field3")
        assert None != sub_config
        assert 456 == sub_config.get("field31")

        config.put("field3.field32", "XYZ")
        assert "XYZ" == config.get("field3.field32")

    def test_from_config(self):
        config = ConfigParams.from_tuples(
            "field1.field11", 123,
            "field2", "ABC",
            "field1.field12", "XYZ"
        )

        params = Parameters.from_config(config)
        assert 2 == len(params)
        assert "ABC" == params.get("field2")
        value = params.get_as_map("field1")
        assert 2 == len(value)
        assert "123" == value.get("field11")
        assert "XYZ" == value.get("field12")



"""
Pip.Services Components
--------------------

Pip.Services is an open-source library of basic microservices.
pip_services4_components package provides basic abstractions portable across variety of languages.

Links
`````

* `website <http://github.com/pip-services/pip-services>`
* `development version <https://github.com/pip-services4/pip-services4-python/tree/main/pip-services4-components-python>`

"""

from setuptools import setup
from setuptools import find_packages

try:
    readme = open('README.md').read()
except:
    readme = __doc__

setup(
    name='pip_services4_components',
    version='0.0.9',
    url='https://github.com/pip-services4/pip-services4-python/tree/main/pip-services4-components-python',
    license='MIT',
    description='Component definitions for Pip.Services in Python',
    author='Conceptual Vision Consulting LLC',
    author_email='seroukhov@gmail.com',
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=['config', 'data', 'test']),
    include_package_data=True,
    zip_safe=True,
    platforms='any',
    install_requires=[
        'PyYAML >= 6.0, < 7.0',
        'pip_services4_commons >= 0.0.1, < 1.0', 
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]    
)



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.exec.IParameterized
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Interface for parameterized components
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from abc import ABC

from ..exec.Parameters import Parameters


class IParameterized(ABC):
    """
    Interface for components that require execution parameters.
    """

    def set_parameters(self, parameters: Parameters):
        """
        Sets execution parameters.

        :param parameters: execution parameters.
        """
        raise NotImplementedError('Method from interface definition')



# -*- coding: utf-8 -*-
"""
    pip_services4_components.context.ContextInfo
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Context context implementation

    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

import datetime
import socket
from typing import Any

from pip_services4_commons.data import StringValueMap

from pip_services4_components.config import IReconfigurable, ConfigParams
from pip_services4_components.context import ContextInfo


class ContextInfo(IReconfigurable):
    """
    Context information component that provides detail information
    about execution context: container or/and process.

    Most often ContextInfo is used by logging and performance counters
    to identify source of the collected logs and metrics.

    ### Configuration parameters ###
        - name:                 the context (container or process) name
        - description:          human-readable description of the context
        - properties:           entire section of additional descriptive properties
        - ...

    Example:

    .. code-block:: python

        contextInfo = ContextInfo()
        contextInfo.configure(ConfigParams.from_tuples(
                                "name", "MyMicroservice",
                                "description", "My first microservice"))

        context.name			# Result: "MyMicroservice"
        context.contextId		# Possible result: "mylaptop"
        context.startTime		# Possible result: 2018-01-01:22:12:23.45Z
        context.uptime			# Possible result: 3454345
    """

    def __init__(self, name: str = None, description: str = None):
        """
        Creates a new instance of this context context.

        :param name: (optional) a context name.

        :param description: (optional) a human-readable description of the context.
        """
        self.__name: str = name or "unknown"
        self.__description: str = description
        self.__properties: StringValueMap = StringValueMap()
        self.__context_id = socket.gethostname()
        self.__start_time: datetime.datetime = datetime.datetime.now()
        self.__uptime: float = 0

    def configure(self, config: ConfigParams):
        """
        Configures component by passing configuration parameters.

        :param config: configuration parameters to be set.
        """
        self.__name = config.get_as_string_with_default("name", self.__name)
        self.__name = config.get_as_string_with_default("context.name", self.__name)

        self.__description = config.get_as_string_with_default("description", self.__description)
        self.__description = config.get_as_string_with_default("context.description", self.__description)

        self.__properties = config.get_section("properties")

    @property
    def name(self) -> str:
        """
        Gets the context name.

        :return: the context name
        """
        return self.__name

    @name.setter
    def name(self, name: str):
        """
        Sets the context name.

        :param name: a new name for the context.
        """
        self.__name = name if name is not None else "unknown"

    @property
    def description(self) -> str:
        """
        Gets the human-readable description of the context.

        :return: the human-readable description of the context.
        """
        return self.__description

    @description.setter
    def description(self, description: str):
        """
        Sets the human-readable description of the context.

        :param description: a new human readable description of the context.
        """
        self.__description = description

    @property
    def context_id(self) -> str:
        """
        Gets the unique context id. Usually it is the current host name.

        :return: the unique context id.
        """
        return self.__context_id

    @context_id.setter
    def context_id(self, context_id: str):
        """
        Sets the unique context id.

        :param context_id: a new unique context id.
        """
        self.__context_id = context_id

    @property
    def start_time(self) -> datetime.datetime:
        """
        Gets the context start time.

        :return: the context start time.
        """
        return self.__start_time

    @start_time.setter
    def start_time(self, start_time: datetime.datetime):
        """
        Sets the context start time.

        :param start_time: a new context start time.
        """
        self.__start_time = start_time

    @property
    def uptime(self) -> float:
        """
        Calculates the context uptime as from the start time.

        :return: number of milliseconds from the context start time.
        """
        return self.__uptime

    @uptime.setter
    def uptime(self, uptime):
        self.__uptime = uptime

    @property
    def properties(self) -> Any:
        """
        Gets context additional parameters.

        :return: a JSON object with additional context parameters.
        """
        return self.__properties

    @properties.setter
    def properties(self, properties: Any):
        """
        Sets context additional parameters.

        :param properties: a JSON object with context additional parameters
        """
        self.__properties = properties

    @staticmethod
    def from_config(config: ConfigParams) -> ContextInfo:
        """
        Creates a new ContextInfo and sets its configuration parameters.

        :param config: configuration parameters for the new ContextInfo.

        :return: a newly created ContextInfo
        """
        value = ContextInfo()
        value.configure(config)
        return value



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.exec.INotifiable
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Interface for notifiable components with parameters
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from abc import ABC
from typing import Optional

from pip_services4_components.context.IContext import IContext

from ..exec.Parameters import Parameters


class INotifiable(ABC):
    """
    Interface for components that can be asynchronously notified.
    The notification may include optional argument that describe the occured event_name

    .. code-block:: python
        class MyComponent(INotifable):
            ...
            def notify(context, args):
                print("Occured event " + args.get_as_string("event"))

        my_component = MyComponent()
        my_component.notify(Context.from_trace_id("123"), Parameters.from_tuples("event", "Test Event"));
    """

    def notify(self, context: Optional[IContext], args: Parameters):
        """
        Notifies the component about occured event_name.

        :param context: (optional) transaction id to trace execution through call chain.

        :param args: notification arguments.
        """
        raise NotImplementedError('Method from interface definition')



# -*- coding: utf-8 -*-
"""
    pip_services4_commons.exec.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Contains design patterns for the standard lifecycle of objects (opened,
    closed, openable, closable, runnable). Helper classes for lifecycle provisioning.

    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = [
    'ICleanable', 'Cleaner',
    'IOpenable', 'Opener', 'IClosable', 'Closer',
]

from .Cleaner import Cleaner
from .Closer import Closer
from .ICleanable import ICleanable
from .IClosable import IClosable
from .IOpenable import IOpenable
from .Opener import Opener



# -*- coding: utf-8 -*-
"""
    pip_services4_components.build.Factory
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Factory implementation
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import List, Any

from .CreateException import CreateException
from .IFactory import IFactory


class Registration:
    def __init__(self, locator, factory):
        self.locator = locator
        self.factory = factory

    locator = None
    factory = None


class Factory(IFactory):
    """
    Basic component factory that creates components using registered types and factory functions.

    Example:

    .. code-block:: python
    
        factory = Factory()

        factory.registerAsType(Descriptor("mygroup", "mycomponent1", "default", "*", "1.0"), MyComponent1)

        factory.create(Descriptor("mygroup", "mycomponent1", "default", "name1", "1.0"))
    """

    def __init__(self):
        self.__registrations: List[Registration] = []

    def register(self, locator: Any, factory: Any):
        """
        Registers a component using a factory method.

        :param locator: a locator to identify component to be created.

        :param factory: a factory function that receives a locator and returns a created component.
        """
        if locator is None:
            raise Exception("Locator cannot be null")
        if factory is None:
            raise Exception("Factory cannot be null")

        self.__registrations.append(Registration(locator, factory))

    def register_as_type(self, locator: Any, component_type: Any):
        """
        Registers a component using its type (a constructor function).

        :param locator: a locator to identify component to be created.

        :param component_type: a component type.
        """
        if locator is None:
            raise Exception("Locator cannot be null")
        if component_type is None:
            raise Exception("Factory cannot be null")

        def factory(loc):
            return component_type()

        self.__registrations.append(Registration(locator, factory))

    def can_create(self, locator: Any) -> Any:
        """
        Checks if this factory is able to create component by given locator.

        This method searches for all registered components and returns
        a locator for component it is able to create that matches the given locator.
        If the factory is not able to create a requested component is returns null.

        :param locator: a locator to identify component to be created.

        :return: a locator for a component that the factory is able to create.
        """
        for registration in self.__registrations:
            this_locator = registration.locator
            if this_locator == locator:
                return this_locator
        return None

    def create(self, locator: Any) -> Any:
        """
        Creates a component identified by given locator.

        :param locator: a locator to identify component to be created.

        :return: the created component.
        """
        for registration in self.__registrations:
            this_locator = registration.locator

            if this_locator == locator:
                try:
                    return registration.factory(locator)
                except Exception as ex:
                    if isinstance(ex, CreateException):
                        raise ex

                    raise CreateException(
                        None,
                        "Failed to create object for " + str(locator)
                    ).with_cause(ex)
